use std::f32::consts::PI;
use std::path::PathBuf;
use std::time::Instant;

use cgmath::{Deg, Euler, Point3, Quaternion, Rad};
use imgui::FontSource;
use legion::systems::CommandBuffer;
use legion::world::SubWorld;
use legion::IntoQuery;
use legion::*;
use nalgebra::Quaternion as naQuaternion;
use rapier3d::dynamics::{
    IntegrationParameters, JointSet, MassProperties, RigidBodyBuilder, RigidBodySet,
};
use rapier3d::geometry::{BroadPhase, ColliderBuilder, ColliderSet, NarrowPhase};
use rapier3d::na::{Isometry3, Vector, Vector3};
use rapier3d::pipeline::{ChannelEventCollector, PhysicsPipeline};

use crate::camera::{Camera, CameraController};
use crate::components::{Collider, LoopState, Mesh, Physics, Position};
use crate::geometry::RawMesh;
use crate::physics::state::PhysicsState;
use crate::render::state::RenderState;
use crate::runtime::state::RuntimeState;

#[system]
#[write_component(Mesh)]
pub fn runtime_load(world: &mut SubWorld, #[resource] runtime_state: &mut RuntimeState) {
    runtime_state.preload_meshes(PathBuf::from("./resources"));
}

#[system]
#[write_component(Mesh)]
pub fn runtime_spawn(
    cmd: &mut CommandBuffer,
    world: &mut SubWorld,
    #[resource] runtime_state: &mut RuntimeState,
    #[resource] renderer: &mut RenderState,
) {
    for entity in &runtime_state.get_entities() {
        match entity.type_name.as_ref() {
            "PhysicsEntity" => {
                let mesh_name = entity.mesh.as_ref().unwrap();
                let raw_mesh = match runtime_state.get_mesh(mesh_name.as_str()) {
                    None => {
                        log::warn!("Skipping entity with invalid mesh file {:?} ", mesh_name);
                        continue;
                    }
                    Some(mesh) => mesh,
                };

                let collider = ColliderBuilder::trimesh(
                    raw_mesh.vertices.iter().map(|v| v.position()).collect(),
                    raw_mesh.indices.clone(),
                )
                .build();

                let collider = ColliderBuilder::capsule_y(2.0, 1.0).build();

                let gpu_mesh_buffer = renderer
                    .upload_mesh_to_buffer(
                        raw_mesh,
                        Some(wgpu::Color {
                            r: 1.0,
                            g: 0.7,
                            b: 0.3,
                            a: 1.0,
                        }),
                    )
                    .unwrap();

                let position = Position::from(entity.position.clone());

                let mut dynamic_body = RigidBodyBuilder::new_dynamic()
                    .can_sleep(false)
                    .mass(100.0)
                    .translation(position.x, position.y, position.z)
                    .build();

                let entity: Entity = cmd.push((
                    position,
                    gpu_mesh_buffer,
                    Physics {
                        rigid_body: dynamic_body,
                        rigid_body_handle: None,
                    },
                    Collider {
                        collider: collider,
                        collider_handle: None,
                    },
                ));
            }
            "Terrain" => {
                let mesh_name = entity.mesh.clone().unwrap();
                let raw_mesh = match runtime_state.get_mesh(mesh_name.as_str()) {
                    None => {
                        log::warn!("Skipping entity with invalid mesh file {:?} ", mesh_name);
                        continue;
                    }
                    Some(mesh) => mesh,
                };

                let position = Position::from(entity.position.clone());

                let mut static_body = RigidBodyBuilder::new_static()
                    .position(Isometry3::new(
                        Vector3::new(position.x, position.y, position.z),
                        Vector::y(),
                    ))
                    .build();

                let mesh_collider = ColliderBuilder::trimesh(
                    raw_mesh.vertices.iter().map(|v| v.position()).collect(),
                    raw_mesh.indices.clone(),
                )
                .build();

                let gpu_mesh_buffer = renderer
                    .upload_mesh_to_buffer(
                        raw_mesh,
                        Some(wgpu::Color {
                            r: 1.0,
                            g: 0.7,
                            b: 0.3,
                            a: 1.0,
                        }),
                    )
                    .unwrap();

                let entity: Entity = cmd.push((
                    position,
                    gpu_mesh_buffer,
                    Physics {
                        rigid_body: static_body,
                        rigid_body_handle: None,
                    },
                    Collider {
                        collider: mesh_collider,
                        collider_handle: None,
                    },
                ));
            }
            "Camera" => {
                let position = Position::from(entity.position.clone());
                let entity: Entity = cmd.push((
                    Camera {
                        position: cgmath::Point3 {
                            x: position.x,
                            y: position.y,
                            z: position.z,
                        },
                        yaw: Rad(-PI),
                        pitch: Rad(PI / 2.0),
                    },
                    CameraController::new(5.0, 1.0),
                ));
            }
            "Light" => {
                let mesh_name = entity.mesh.clone().unwrap();
                let raw_mesh = match runtime_state.get_mesh(mesh_name.as_str()) {
                    None => {
                        log::warn!("Skipping entity with invalid mesh file {:?} ", mesh_name);
                        continue;
                    }
                    Some(mesh) => mesh,
                };

                let position = Position::from(entity.position.clone());

                let gpu_mesh_buffer = renderer
                    .upload_mesh_to_buffer(
                        raw_mesh,
                        Some(wgpu::Color {
                            r: 1.0,
                            g: 0.7,
                            b: 0.3,
                            a: 1.0,
                        }),
                    )
                    .unwrap();

                let light_entity: Entity =
                    cmd.push((position, gpu_mesh_buffer, renderer.create_light()));
            }

            _ => {}
        }
    }
}
